<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="2.0"
    xmlns:cam2func="http://jcam.org.uk/editor/function"
    xmlns:as="http://www.oasis-open.org/committees/cam"
    xmlns:look="http://jcam.org.uk/LookupLists"
    xmlns:camed="http://jcam.org.uk/editor">
    <!-- Provided "as is" License: Creative Commons - http://creativecommons.org/licenses/by-sa/3.0/
        Authors: David RR Webber, January, 2009 -->
    <!--
        Purpose  : Parse CAM cxf file and rename the elements and attributes.
                   Inputs: Mode parameter - can be NIEM or CamelCase.
        Original : 01/01/2009
        Initials : drrw
        Description : Parse exchange template and for each element and attribute - inspect and correct naming conventions used.
                              Selectable by namespace and renaming domain mode.
                              Note - does NOT change the root node element name!
                              Note - currently does not change names inside choice conditions - contains() statements
       
        Updated  - Inits - Change
        06/06/09   drrw   Initial release 
        06/10/09   drrw   Fix file extension to xml for xref
        08/21/09   drrw   Support morphing name elements to NIEM person name types
        12/23/09   drrw   Add parameter for method - stripNS, replaceNS
        12/23/09   drrw   Add parameter for namespace insertion
        01/11/10   drrw   Add next word parsing to name logic
        01/14/10   drrw   Add refactor mode support
        01/14/10   drrw   Add namespace remove and strip modes
        01/14/10   drrw   Rename report enhancements
        01/14/10   drrw   Renaming - split words on case change
        01/21/10   drrw   New add namespace prefix mode
        01/21/10   drrw   Add ability to fix NIEM term rules matching
        01/22/10   drrw   Add list of 200 most common mis-spellings from internet websites text
                                    http://en.wikipedia.org/wiki/Commonly_misspelled_words#Documented_list_of_common_misspellings
        01/22/10  drrw    Support ~ prefix for target namespace not to rename in targetNS
        01/23/10  drrw    NDR term name refactoring
        01/25/10  drrw    Refactoring remove nillable rules
        02/01/10  drrw    Refactor add missing datetime, date, time and number rules from Type annotation
        02/01/10  drrw    Hints for enumeration restrictValues via annotation Values tab
        02/01/10  drrw    Annotation Rename tags for spelling words
        02/05/10  drrw    Fix for empty annotations list
        03/05/10  drrw    Insert namespace prefix logic for all XPaths
        04/07/10  drrw    CEN abbreviations
        06/16/10  drrw    More abbreviations in word utils
        06/16/10  drrw    Next name parameter fix
        01/18/11  drrw   In NIEM mode - model names correcting of mixed case named items containing uppercase type name
        01/23/11  drrw   Output as:Extension section
        06/18/11  drrw   Check for lowercase representation terms at end of element names
    -->
    <xsl:param name="resultNamePath">new-template.cxf</xsl:param>
    <xsl:param name="renameMode">NIEM</xsl:param><!-- NIEM or camel case -->
    <xsl:param name="targetNS"></xsl:param>
    <xsl:param name="renameMethod">normal</xsl:param><!-- normal, replace namespace, remove namespace -->
    <xsl:param name="refactorMode">true</xsl:param>
    <xsl:param name="insertNS"></xsl:param><!-- ghy: dhs: oasis: etc, with or without : suffix works -->

    <xsl:output name="resultXML" method="xml" indent="yes" omit-xml-declaration="yes" />
    
    <xsl:variable name="renameXref">
        <xsl:choose>
            <xsl:when test="contains($resultNamePath,'.')"><xsl:value-of select="concat(substring-before($resultNamePath,'.'),'-crossref.xml')"></xsl:value-of></xsl:when>
            <xsl:otherwise><xsl:value-of select="concat($resultNamePath,'-crossref.xml')"/></xsl:otherwise>
        </xsl:choose>
    </xsl:variable>
    
    <xsl:variable name="version">1.16</xsl:variable>
    <xsl:variable name="nlchar" select="'&#xA;'"/>
    <xsl:variable name="rootElement" select="as:CAM/as:AssemblyStructure/as:Structure/as:Element/@name[1]"/>
    <xsl:variable name="rootElementId" select="as:CAM/as:AssemblyStructure/as:Structure/as:Element[@name=$rootElement]/as:Attribute[@name='Id']/@exampleValue"/>
    <xsl:variable name="insertNSprefix" select="if($insertNS='') then '' else if(ends-with($insertNS,':')) then $insertNS else concat($insertNS,':')"></xsl:variable>
    
    <xsl:variable name="source" select="/"/>
    <xsl:variable name="rules" select="/"/>  <!-- Used by the NDR functions -->  
    <xsl:variable name="ignoreDomainsNDR" select="if($renameMode='NIEM') then 'false' else 'true'"/>

    <xsl:include href="word-utils.xsl"/>
        
    <xsl:include href="cam-NDR-functions.xsl"/>
    
      <xsl:template match="/">

<xsl:message>
    Renamer utility, version <xsl:value-of select="$version"/>
    Rename mode    - <xsl:value-of select="$renameMode"/>
    Refactor mode    - <xsl:value-of select="$refactorMode"/>
    Target namespace - <xsl:value-of select="$targetNS"/>
     <xsl:if test="not($insertNS='')">Inserting namespace - <xsl:value-of select="$insertNS"/></xsl:if> 
     Writing output to - <xsl:value-of select="$resultNamePath"/>
     Generated on     - <xsl:value-of select="current-dateTime()"></xsl:value-of></xsl:message>

          <xsl:variable name="allElements">
              <allElements>
                  <xsl:apply-templates select="as:CAM/as:AssemblyStructure/as:Structure/as:Element/*" mode="list"><xsl:with-param name="type">element</xsl:with-param></xsl:apply-templates> 
              </allElements>
          </xsl:variable>
          <xsl:variable name="allAttributes">
              <allAttributes>
                  <xsl:apply-templates select="as:CAM/as:AssemblyStructure/as:Structure/as:Element/*" mode="list"><xsl:with-param name="type">attribute</xsl:with-param></xsl:apply-templates> 
              </allAttributes>
          </xsl:variable>
          
          <xsl:message> Renaming started</xsl:message>
          <xsl:message> Processing <xsl:value-of select="count(distinct-values($allElements/allElements//element/@name))"/> elements </xsl:message>
          <xsl:variable name="renamedElements">
              <allElements>
                  <xsl:for-each select="distinct-values($allElements/allElements//element/@name)">
                      <xsl:variable name="thisName" select="."/>
                      <xsl:variable name="chkName" select="if(contains(.,':')) then substring-after(.,':') else ."/>
                      <xsl:variable name="ns" select="if(contains(.,':')) then concat(substring-before(.,':'),':') else ''"/>
                      <xsl:variable name="firstLetter" select="substring($chkName,1,1)"/>
                     
                      <xsl:variable name="newname">
                          <xsl:choose>
                              <xsl:when test="$renameMode='NIEM'">
                                  <xsl:variable name="typeNameList" select="$source/as:CAM/as:AssemblyStructure/as:Structure//as:Element[@name=$thisName]/camed:annotation/camed:documentation[@type='Type']"/>
                                  <xsl:variable name="typeName" select="$typeNameList[1]"/>
                              
                                  <xsl:variable name="adjustMixedName"><xsl:if test="string-length($typeName) &gt; 0 and(starts-with($thisName,normalize-space($typeName))) and(not(upper-case($thisName) = $thisName))">true</xsl:if></xsl:variable>
                                  <!-- xsl:message>This:  <xsl:value-of select="$thisName"/> Type: <xsl:value-of select="$typeName"/> - <xsl:value-of select="$adjustMixedName"/> - <xsl:value-of select="substring($thisName,1,string-length($typeName)) = $typeName"/> - <xsl:value-of select="string-length(normalize-space($typeName))"/></xsl:message -->
                                  <xsl:choose>
                                      <xsl:when test="$adjustMixedName='true'"><xsl:value-of select="concat($ns,camed:FIX-UPPER(upper-case($chkName),$renameMode,'ELEMENT'))"/></xsl:when>
                                      <xsl:when test="camed:WordMatch($chkName,$renameMode)='true'"></xsl:when>
                                      <xsl:when test="$firstLetter=$chkName and(upper-case($firstLetter)=$firstLetter)"></xsl:when><!-- skip one char upper case name -->
                                      <xsl:when test="upper-case($chkName)=$chkName"><xsl:value-of select="concat($ns,camed:FIX-UPPER($chkName,$renameMode,'ELEMENT'))"/></xsl:when>                                      
                                      <xsl:when test="not(upper-case($firstLetter)=$firstLetter)"><xsl:value-of select="concat($ns,replace(replace(concat(upper-case($firstLetter),substring-after($chkName,$firstLetter)),'_','-'),'\.','-'))"/></xsl:when><!-- Make first letter upper case -->
                                      <xsl:when test="contains($chkName,'_') or(contains($chkName,'.'))"><xsl:value-of select="concat($ns,replace(replace($chkName,'_','-'),'\.','-'))"/></xsl:when><!-- Adjust special characters -->
                                      <xsl:otherwise/>
                                  </xsl:choose>
                              </xsl:when>
                              <xsl:when test="$renameMode='CamelCase'">
                                  <xsl:choose>
                                      <xsl:when test="camed:WordMatch($chkName,$renameMode)='true'"/>
                                      <xsl:when test="$firstLetter=$chkName and(upper-case($firstLetter)=$firstLetter)"/><!-- skip one char upper case name -->
                                      <xsl:when test="upper-case($chkName)=$chkName"><xsl:value-of select="concat($ns,camed:FIX-UPPER($chkName,$renameMode,'ELEMENT'))"/></xsl:when>
                                      <xsl:when test="not(upper-case($firstLetter)=$firstLetter)"><xsl:value-of select="concat($ns,upper-case($firstLetter),substring-after($chkName,$firstLetter))"/></xsl:when><!-- Make first letter upper case -->
                                      <xsl:otherwise/>
                                  </xsl:choose>
                              </xsl:when>
                              <xsl:otherwise/>
                          </xsl:choose>
                      </xsl:variable>
                      <xsl:if test="string-length($newname) &gt; 0 and(not($newname=$ns))">                          
                          <!-- Compute new element name -->
                          <xsl:variable name="completeNewName">
                             <xsl:choose>
                                 <xsl:when test="$renameMode='NIEM' and($refactorMode='true')">
                                     <!-- xsl:when test=""></xsl:when --><!-- ??? NIEM name property mode with property and mixed case -->
                                     <!-- Now check NDR term rules on name and fix accordingly -->
                                     <xsl:variable name="chkName" select="cam2func:has-representation-term($newname)"/>
                                     <!-- xsl:message> - chkname - <xsl:value-of select="$chkName"/></xsl:message -->
                                     <xsl:choose>
                                         <xsl:when test="string-length($chkName) &gt; 0">
                                             <!-- Now correct the representation term based on content type -->
                                             <xsl:choose>
                                                 <xsl:when test="$source//as:Element[@name=$thisName]/as:Element and(not($source//as:Element[@name=$thisName]/as:Element[@name=$thisName]))"><xsl:value-of select="$newname"/></xsl:when><!-- Ignore parent elements -->
                                                 <xsl:otherwise> 
                                                     <xsl:variable name="baseDataTypes">
                                                         <xsl:for-each select="$source//as:Element[@name=$thisName][1]/as:Rule">
                                                             <!-- check the constraint details -->
                                                             <xsl:for-each select="as:constraint">
                                                                 <xsl:value-of select="cam2func:rules2type(@action)"/>
                                                             </xsl:for-each>
                                                         </xsl:for-each>
                                                     </xsl:variable>
                                                     <xsl:variable name="baseDataType" select="distinct-values(tokenize($baseDataTypes,'\|'))"/>
                                                     <xsl:variable name="baseTypesCount" select="count($baseDataType)"/>
                                                     <xsl:variable name="NameSuffix">
                                                         <xsl:for-each select="$baseDataType">
                                                             <xsl:if test="$baseTypesCount &lt; 3 or(not(.='string'))"><xsl:value-of select="cam2func:type2term(.)"/></xsl:if>
                                                         </xsl:for-each>
                                                     </xsl:variable>
                                                     <xsl:choose>
                                                         <xsl:when test="ends-with(upper-case($newname),upper-case($NameSuffix))">
                                                             <xsl:value-of  select="concat(substring($newname,1,string-length($newname) - string-length($NameSuffix)),$NameSuffix)"/>
                                                         </xsl:when>
                                                         <xsl:otherwise><xsl:value-of select="concat($newname,$NameSuffix)"/></xsl:otherwise>
                                                     </xsl:choose>
                                                     <!-- xsl:message>Fix: <xsl:value-of select="$newname"/><xsl:value-of select="$NameSuffix"/> | <xsl:value-of select="$baseDataType"/></xsl:message -->
                                                 </xsl:otherwise>
                                             </xsl:choose>
                                         </xsl:when>
                                         <xsl:otherwise><xsl:value-of select="$newname"/></xsl:otherwise>
                                     </xsl:choose>
                                 </xsl:when>
                                 <xsl:otherwise><xsl:value-of select="$newname"/></xsl:otherwise>
                             </xsl:choose>
                          </xsl:variable>
                          <xsl:variable name="newnameValue" select="if(contains($completeNewName,':')) then substring-after($completeNewName,':') else $completeNewName"/>
                          <xsl:variable name="splitNewName"><xsl:value-of select="camed:splitNewWords($newnameValue,$newnameValue,'true')"/></xsl:variable>
                          <xsl:variable name="spellChkName">
                              <xsl:for-each select="tokenize($splitNewName,'\|')"><xsl:if test="not(.='')"><xsl:value-of  select="concat(.,' ')"/></xsl:if></xsl:for-each>
                          </xsl:variable>
                          <!-- Now write renaming details --> 
                          <element name='{.}' newname='{$completeNewName}' spellcheck='{$spellChkName}' origname='{$newname}'/>
                      </xsl:if>
              </xsl:for-each>
              </allElements>
          </xsl:variable>
          
          <xsl:variable name="renamedAttributes">
              <allAttributes>
                  <xsl:for-each select="distinct-values($allAttributes/allAttributes//attribute/@name)">
                      <xsl:variable name="thisName" select="."/>
                      <xsl:variable name="chkName" select="if(contains(.,':')) then substring-after(.,':') else ."/>
                      <xsl:variable name="ns" select="if(contains(.,':')) then concat(substring-before(.,':'),':') else ''"/>
                      <xsl:variable name="firstLetter" select="substring($chkName,1,1)"/>
                      <xsl:variable name="newname">
                          <xsl:choose>
                              <xsl:when test="$renameMode='NIEM'">
                                  <xsl:choose>
                                      <xsl:when test="camed:WordMatch($chkName,$renameMode)='true'"/>
                                      <xsl:when test="upper-case($chkName)=$chkName and(string-length($chkName) &gt; 1)"><xsl:value-of select="concat($ns,camed:FIX-UPPER($chkName,$renameMode,'ATTRIBUTE'))"/></xsl:when>                                    
                                      <xsl:when test="upper-case($firstLetter)=$firstLetter"><xsl:value-of select="concat($ns,replace(replace(concat(lower-case($firstLetter),substring-after($chkName,$firstLetter)),'_','-'),'\.','-'))"/></xsl:when><!-- Make first letter lower case -->
                                      <xsl:when test="contains($chkName,'_') or(contains($chkName,'.'))"><xsl:value-of select="concat($ns,replace(replace($chkName,'_','-'),'\.','-'))"/></xsl:when><!-- Adjust special characters -->
                                      <xsl:otherwise/>
                                  </xsl:choose>
                              </xsl:when>
                              <xsl:when test="$renameMode='CamelCase'">
                                  <xsl:choose>
                                      <xsl:when test="camed:WordMatch($chkName,$renameMode)='true'"/>
                                      <xsl:when test="upper-case($chkName)=$chkName and(string-length($chkName) &gt; 1)"><xsl:value-of select="concat($ns,camed:FIX-UPPER($chkName,$renameMode,'ATTRIBUTE'))"/></xsl:when>
                                      <xsl:when test="not(upper-case($firstLetter)=$firstLetter)"><xsl:value-of select="concat($ns,upper-case($firstLetter),substring-after($chkName,$firstLetter))"/></xsl:when><!-- Make first letter upper case -->
                                      <xsl:otherwise/>
                                  </xsl:choose>
                              </xsl:when>
                              <xsl:otherwise/>
                          </xsl:choose>
                      </xsl:variable>
                      <xsl:if test="string-length($newname) &gt; 0 and(not($newname=$ns))">
                          <!-- compute attribute name -->
                          <xsl:variable name="completeNewName">
                              <xsl:choose>
                                  <xsl:when test="$renameMode='NIEM' and($refactorMode='true')">
                                      <!-- Now check NDR term rules on name and fix accordingly -->
                                      <xsl:variable name="chkName" select="cam2func:has-representation-term($newname)"/>
                                      <xsl:choose>
                                          <xsl:when test="string-length($chkName) &gt; 0">
                                              <!-- Now correct the representation term based on content type -->
                                                      <xsl:variable name="baseDataTypes">
                                                          <xsl:for-each select="$source//as:Attribute[@name=$thisName][1]/as:Rule">
                                                              <!-- check the constraint details -->
                                                              <xsl:for-each select="as:constraint">
                                                                  <xsl:value-of select="cam2func:rules2type(@action)"/>
                                                              </xsl:for-each>
                                                          </xsl:for-each>
                                                      </xsl:variable>
                                                      <xsl:variable name="baseDataType" select="distinct-values(tokenize($baseDataTypes,'\|'))"/>
                                                      <xsl:variable name="NameSuffix">
                                                          <xsl:for-each select="$baseDataType">
                                                              <xsl:value-of select="cam2func:type2term(.)"/>
                                                          </xsl:for-each>
                                                      </xsl:variable>
                                                    <xsl:choose>
                                                        <xsl:when test="ends-with(upper-case($newname),upper-case($NameSuffix))">
                                                            <xsl:value-of  select="concat(substring($newname,1,string-length($newname) - string-length($NameSuffix)),$NameSuffix)"/>
                                                        </xsl:when>
                                                        <xsl:otherwise><xsl:value-of select="concat($newname,$NameSuffix)"/></xsl:otherwise>
                                                    </xsl:choose>
                                                      <!-- xsl:message>Fix attr: <xsl:value-of select="$newname"/><xsl:value-of select="$NameSuffix"/> | <xsl:value-of select="$baseDataType"/></xsl:message -->
                                          </xsl:when>
                                          <xsl:otherwise><xsl:value-of select="$newname"/></xsl:otherwise>
                                      </xsl:choose>
                                  </xsl:when>
                                  <xsl:otherwise><xsl:value-of select="$newname"/></xsl:otherwise>
                              </xsl:choose>
                          </xsl:variable>
                          
                          <xsl:variable name="newnameValue" select="if(contains($completeNewName,':')) then substring-after($completeNewName,':') else $completeNewName"/>
                          <xsl:variable name="splitNewName"><xsl:value-of select="camed:splitNewWords($newnameValue,$newnameValue,'true')"/></xsl:variable>
                          <xsl:variable name="spellChkName">
                              <xsl:for-each select="tokenize($splitNewName,'\|')"><xsl:if test="not(.='')"><xsl:value-of  select="concat(.,' ')"/></xsl:if></xsl:for-each>
                          </xsl:variable>
                          
                          <attribute name='{.}' newname='{$completeNewName}' spellcheck='{$spellChkName}'/>
                      </xsl:if>
                  </xsl:for-each>
              </allAttributes>
          </xsl:variable>

          <xsl:message> Writing template results</xsl:message>

          <xsl:result-document exclude-result-prefixes="as camed" format="resultXML" indent="yes" href="{$resultNamePath}">
            
            <xsl:variable name="results">
            <xsl:comment>
                Renamer utility, version <xsl:value-of select="$version"/>
                Rename mode    - <xsl:value-of select="$renameMode"/>
                Refactor mode    - <xsl:value-of select="$refactorMode"/>
                Target namespace - <xsl:value-of select="$targetNS"/> 
                <xsl:if test="not($insertNS='')">Inserting namespace - <xsl:value-of select="$insertNS"/></xsl:if> 
                Writing output to - <xsl:value-of select="$resultNamePath"/>
                Generated on     - <xsl:value-of select="current-dateTime()"></xsl:value-of>
                <xsl:value-of select="'&#xD; &#xD;' "/>
            </xsl:comment>
                
             <xsl:element name="as:CAM">
                 <xsl:attribute name="CAMlevel" select="as:CAM/@CAMlevel"/>
                 <xsl:attribute name="version" select="as:CAM/@version"/>
                 
                <xsl:apply-templates select="as:CAM/as:Header"/>
                <xsl:apply-templates select="as:CAM/as:Namespaces"/>
                    
                <xsl:apply-templates select="as:CAM/as:AssemblyStructure" mode="rename">
                    <xsl:with-param name="newElements" select="$renamedElements"/>
                    <xsl:with-param name="newAttributes" select="$renamedAttributes"/>
                </xsl:apply-templates>
                 
                 <xsl:copy-of select="as:CAM/as:Extension"/>     <!-- Copy lookups -->
     
             </xsl:element>
            </xsl:variable>
              
              <xsl:for-each select="$results/*"><!-- re-load the namespaces from source CAM -->
                  <xsl:copy>
                      <xsl:for-each select=".//namespace::*">
                          <xsl:if test="not(name()='w2e')">
                              <xsl:namespace name="{name()}" select="."/>
                          </xsl:if>
                      </xsl:for-each>
                      <xsl:copy-of select="@*|node()"/>
                  </xsl:copy>
              </xsl:for-each>
              
          </xsl:result-document>
          <xsl:message> Template results complete</xsl:message>
          <xsl:message> Writing cross-reference</xsl:message>

        <xsl:result-document exclude-result-prefixes="as camed" format="resultXML" indent="yes" href="{$renameXref}">
            
            <RenamingXref>               
                <xsl:comment>
                    Renamer utility, version <xsl:value-of select="$version"/>
                    Rename mode    - <xsl:value-of select="$renameMode"/> 
                    Writing output to - <xsl:value-of select="$resultNamePath"/>
                    For namespace - <xsl:value-of select="$targetNS"/>
                    <xsl:if test="not($insertNS='')">Inserting namespace - <xsl:value-of select="$insertNS"/></xsl:if>
                    Generated on - <xsl:value-of select="current-dateTime()"></xsl:value-of>
                    <xsl:value-of select="'&#xD; &#xD;' "/>
                </xsl:comment>
                
                <xsl:copy-of select="$renamedElements"/>
                <xsl:copy-of select="$renamedAttributes"/>
                
            </RenamingXref>
            
        </xsl:result-document>

        <xsl:message> Renaming finished</xsl:message>
        
    </xsl:template>
    
    <xsl:template match="as:Element" mode="list">
        <xsl:param name="type"/>
        <xsl:variable name="thisNS" select="substring-before(@name,':')"/>
        <xsl:if test="$type='element' and($targetNS='' or($targetNS='*') or($targetNS=$thisNS)) or(starts-with($targetNS,'~') and not($targetNS=concat('~',$thisNS)))">
        <xsl:element name="element">
            <xsl:attribute name="name" select="@name"/>
        </xsl:element>
        </xsl:if>
        <xsl:if test="$type='attribute'">
            <xsl:apply-templates mode="list" select="as:Attribute"><xsl:with-param name="type" select="$type"/></xsl:apply-templates>
        </xsl:if>
        <xsl:apply-templates mode="list" select="as:Element"><xsl:with-param name="type" select="$type"/></xsl:apply-templates>
    </xsl:template>
    
    <xsl:template match="as:Attribute" mode="list">
        <xsl:param name="type"/>
        <xsl:variable name="thisNS" select="substring-before(@name,':')"/>
        <xsl:if test="$type='attribute' and($targetNS='' or($targetNS='*') or($targetNS=$thisNS)) or(starts-with($targetNS,'~') and not($targetNS=concat('~',$thisNS)))">
        <xsl:element name="attribute">
            <xsl:attribute name="name" select="@name"/>
        </xsl:element>
        </xsl:if>
    </xsl:template>
    
    <xsl:template match="node()" mode="list"></xsl:template>
       
    <xsl:template match="comment()">
        <xsl:param name="parentPath"/>
        <xsl:param name="mode"/>
        <xsl:comment><xsl:value-of select="."/></xsl:comment>
    </xsl:template>
    
    <xsl:template match="as:AssemblyStructure" mode="rename">
        <xsl:param name="newElements"/>
        <xsl:param name="newAttributes"/>
            <xsl:element name="as:AssemblyStructure">
                <xsl:for-each select="attribute::*"><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:for-each>
                <xsl:apply-templates select="*" mode="rename">
                    <xsl:with-param name="newElements" select="$newElements"/>
                    <xsl:with-param name="newAttributes" select="$newAttributes"/>
                </xsl:apply-templates>
            </xsl:element>
    </xsl:template>
    
    <xsl:template match="as:Structure" mode="rename">
        <xsl:param name="newElements"/>
        <xsl:param name="newAttributes"/>
        <xsl:element name="as:Structure">
            <xsl:for-each select="attribute::*"><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:for-each>
            <xsl:apply-templates select="*" mode="rename">
                <xsl:with-param name="newElements" select="$newElements"/>
                <xsl:with-param name="newAttributes" select="$newAttributes"/>
            </xsl:apply-templates>
        </xsl:element>
    </xsl:template>
    
    <xsl:template match="as:Element" mode="rename">
        <xsl:param name="newElements"/>
        <xsl:param name="newAttributes"/>

        <xsl:variable name="this" select="@name"/>
        <xsl:variable name="newname" select="$newElements//allElements/element[@name=$this]/@newname"/>
        <xsl:variable name="prefix" select="if(contains($this,':') or(contains($newname,':'))) then '' else $insertNSprefix"/>

        <xsl:element name="as:Element">
            <xsl:choose>
                <xsl:when test="string-length($newname) &gt; 0">
                        <xsl:for-each select="attribute::*">
                            <xsl:choose>
                                <xsl:when test="name()='name'"><xsl:attribute name="{name()}"><xsl:value-of select="concat($prefix,$newname)"/></xsl:attribute></xsl:when><!-- replace name with new name -->
                                <xsl:when test="name()='path'"><xsl:attribute name="{name()}"><xsl:value-of select="camed:chkPath(.,$newElements,'')"/></xsl:attribute></xsl:when><!-- lookup path changes -->
                                <xsl:when test="name()='namespacePrefix' and(not($prefix=''))"><xsl:attribute name="{name()}"><xsl:value-of select="replace($prefix,':','')"/></xsl:attribute></xsl:when><!-- insert namespace -->
                                <xsl:when test="name()='namespace' and(not($prefix=''))"><xsl:attribute name="{name()}"><xsl:value-of select="$source/as:CAM/as:Namespaces/as:namespace[@prefix=replace($prefix,':','')]/text()"/></xsl:attribute></xsl:when><!-- insert namespace -->
                                <xsl:otherwise><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:otherwise>
                            </xsl:choose>
                        </xsl:for-each>
                </xsl:when>
                <xsl:otherwise>              
                        <xsl:for-each select="attribute::*">
                            <xsl:choose>
                                <xsl:when test="name()='name'"><xsl:attribute name="{name()}"><xsl:value-of select="concat($prefix,.)"/></xsl:attribute></xsl:when><!-- use existing name --> 
                                <xsl:when test="name()='path'"><xsl:attribute name="{name()}"><xsl:value-of select="camed:chkPath(.,$newElements,'')"/></xsl:attribute></xsl:when><!-- lookup path changes -->
                                <xsl:when test="name()='namespacePrefix' and(not($prefix=''))"><xsl:attribute name="{name()}"><xsl:value-of select="replace($prefix,':','')"/></xsl:attribute></xsl:when><!-- insert namespace -->
                                <xsl:when test="name()='namespace' and(not($prefix=''))"><xsl:attribute name="{name()}"><xsl:value-of select="$source/as:CAM/as:Namespaces/as:namespace[@prefix=replace($prefix,':','')]/text()"/></xsl:attribute></xsl:when><!-- insert namespace -->
                                <xsl:otherwise><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:otherwise>
                            </xsl:choose>
                        </xsl:for-each>
                </xsl:otherwise>
            </xsl:choose>
            
            <xsl:variable name="newRuleChanges">
                <xsl:if test="count(as:Element)=0 and($refactorMode='true')"><!-- Leaf node and refactoring on -->
                    <xsl:choose>
                        <xsl:when test="ends-with($newname,'Time') and(not(as:Rule/as:constraint[contains(@action,'setDateMask')]))">
                            <xsl:copy-of  select="camed:annotation"/>
                        </xsl:when>
                        <xsl:when test="ends-with($newname,'Date') and(not(as:Rule/as:constraint[contains(@action,'setDateMask')]))">
                            <xsl:copy-of  select="camed:annotation"/>
                        </xsl:when>
                        <xsl:when test="ends-with($newname,'Flag') and(not(as:Rule/as:constraint[contains(@action,'restrictValues')]))">
                            <xsl:copy-of  select="camed:annotation"/>
                        </xsl:when>
                        <xsl:when test="ends-with($newname,'Code') and(not(contains($newname,'ZIP'))) and(not(as:Rule/as:constraint[contains(@action,'restrictValues')]))">
                            <xsl:copy-of  select="camed:annotation"/>
                        </xsl:when>
                        <xsl:when test="ends-with($newname,'Type') and(not(as:Rule/as:constraint[contains(@action,'restrictValues')]))">
                            <xsl:copy-of  select="camed:annotation"/>
                        </xsl:when>
                        <xsl:when test="ends-with($newname,'Indicator') and(not(as:Rule/as:constraint[contains(@action,'restrictValues')]))">
                            <xsl:copy-of  select="camed:annotation"/>
                        </xsl:when>
                        <xsl:otherwise/>
                    </xsl:choose>
                </xsl:if>
            </xsl:variable>  
            
            <!-- xsl:if test="count($newRuleChanges/camed:annotation) &gt; 0"><xsl:message>check rules: <xsl:value-of select="$newname"/> - <xsl:value-of select="count(as:Rule)"/> - ERROR</xsl:message></xsl:if -->
                
            <xsl:apply-templates select="*" mode="rename"><!-- Now process the rules and annotations for the structure -->
                <xsl:with-param name="newElements" select="$newElements"/>
                <xsl:with-param name="newAttributes" select="$newAttributes"/>
                <xsl:with-param name="newRuleChanges" select="$newRuleChanges"/>
            </xsl:apply-templates>
            
        </xsl:element>
    </xsl:template>
    
    <xsl:template match="as:Attribute" mode="rename">
        <xsl:param name="newElements"/>
        <xsl:param name="newAttributes"/>
        
        <xsl:variable name="this" select="@name"/>
        <xsl:variable name="newname" select="$newAttributes//allAttributes/attribute[@name=$this]/@newname"/>
        
        <xsl:choose>
            <xsl:when test="string-length($newname) &gt; 0">
                <xsl:element name="as:Attribute">
                    <xsl:for-each select="attribute::*">
                        <xsl:choose>
                            <xsl:when test="name()='name'"><xsl:attribute name="{name()}"><xsl:value-of select="$newname"/></xsl:attribute></xsl:when><!-- replace name with new name -->
                            <xsl:when test="name()='path'"><xsl:attribute name="{name()}"><xsl:value-of select="camed:chkPath(.,$newElements,$newAttributes)"/></xsl:attribute></xsl:when><!-- lookup path changes -->
                            <xsl:otherwise><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:otherwise>
                        </xsl:choose>
                    </xsl:for-each>
                    <xsl:apply-templates select="*" mode="rename"><!-- Now process annotations and rules -->
                        <xsl:with-param name="newElements" select="$newElements"/>
                        <xsl:with-param name="newAttributes" select="$newAttributes"/>
                    </xsl:apply-templates>
                </xsl:element>
            </xsl:when>
            <xsl:otherwise>
                <xsl:element name="as:Attribute">
                    <xsl:for-each select="attribute::*">
                        <xsl:choose>
                            <xsl:when test="name()='path'"><xsl:attribute name="{name()}"><xsl:value-of select="camed:chkPath(.,$newElements,$newAttributes)"/></xsl:attribute></xsl:when><!-- lookup path changes -->
                            <xsl:otherwise><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:otherwise>
                        </xsl:choose>
                    </xsl:for-each>
                    <xsl:apply-templates select="*" mode="rename"><!-- Now process annotations and rules -->
                        <xsl:with-param name="newElements" select="$newElements"/>
                        <xsl:with-param name="newAttributes" select="$newAttributes"/>
                    </xsl:apply-templates>
                </xsl:element>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    
    <xsl:template match="as:Rule" mode="rename">
        <xsl:param name="newElements"/>
        <xsl:param name="newAttributes"/>
        <xsl:param name="newRuleChanges"/>

        <xsl:element name="as:Rule">
            <xsl:for-each select="as:constraint">
                <xsl:choose>
                    <xsl:when test="$refactorMode='true' and(contains(@action,'makeNillable(xsd)'))"/><!-- Exclude any nillable rules -->
                    <xsl:otherwise>
                        <xsl:element name="as:constraint">
                            <xsl:for-each select="attribute::*">
                                <xsl:choose>
                                    <xsl:when test="name()='xpath'"><xsl:attribute name="{name()}"><xsl:value-of select="if(starts-with(.,'//')) then '/' else ''"/><xsl:value-of select="camed:chkPath(.,$newElements,$newAttributes)"/></xsl:attribute></xsl:when><!-- lookup path changes -->
                                    <xsl:when test="name()='condition'"><xsl:attribute name="{name()}"><xsl:value-of select="camed:chkExpression(.,$newElements,$newAttributes)"/></xsl:attribute></xsl:when><!-- lookup path changes -->
                                    <xsl:otherwise><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:otherwise>
                                </xsl:choose>
                            </xsl:for-each>
                        </xsl:element>
                        <xsl:copy-of select="as:annotation"/>
                    </xsl:otherwise>
                    </xsl:choose>
            </xsl:for-each>
            <!-- Now insert in any new refactoring rules -->
            <xsl:variable name="thisXPath" select='as:constraint/@xpath'/>
            <xsl:if test="not($newRuleChanges='') and(count($newRuleChanges/camed:annotation) &gt; 0)">
             <xsl:if test="$newRuleChanges/camed:annotation/camed:documentation[@type='Type'] 
                 or($newRuleChanges/camed:annotation/camed:documentation[@type='Value'])"> 
                <xsl:message> inserting rule(s): <xsl:value-of select="$newRuleChanges/camed:annotation/@item"/></xsl:message>
                <xsl:for-each select="$newRuleChanges/camed:annotation/camed:documentation[@type='Type']">
                    <xsl:if test="contains(.,'=')">
                        <xsl:variable name="newRule" select="concat(substring-before(.,'='),'(',substring-after(.,'='),')')"/>
                        <!-- rule detail to console log -->
                        <xsl:message> <xsl:value-of select="$newRule"/></xsl:message>

                        <xsl:element name="as:constraint">
                            <xsl:attribute name="action">
                                <xsl:value-of select="$newRule"/>
                            </xsl:attribute>
                            <xsl:attribute name="xpath"><xsl:value-of select="if(starts-with($thisXPath[1],'//')) then '/' else ''"/><xsl:value-of select="camed:chkPath($thisXPath[1],$newElements,$newAttributes)"/></xsl:attribute>
                        </xsl:element>
                    </xsl:if>
                </xsl:for-each>   
                <xsl:for-each select="$newRuleChanges/camed:annotation/camed:documentation[@type='Value']">
                    <xsl:variable name="values" select="tokenize(.,'\|')"/>
                    <xsl:variable name="valuesList">
                        <xsl:for-each select="$values">
                            <xsl:choose>
                                <xsl:when test=".=''"/>
                                <xsl:when test="position()=1"><xsl:value-of select="concat('''',.,'''')"/></xsl:when>
                                <xsl:otherwise> |<xsl:value-of select="concat('''',.,'''')"/></xsl:otherwise>
                            </xsl:choose>
                        </xsl:for-each>
                    </xsl:variable>
                    <xsl:message> <xsl:value-of select="replace(replace($valuesList,' \|','|'),'''','')"/></xsl:message>
                    <xsl:element name="as:constraint">
                        <xsl:attribute name="action">
                            <xsl:value-of select="concat('restrictValues(',$valuesList,')')"/>
                        </xsl:attribute>
                        <xsl:attribute name="xpath"><xsl:value-of select="if(starts-with($thisXPath[1],'//')) then '/' else ''"/><xsl:value-of select="camed:chkPath($thisXPath[1],$newElements,$newAttributes)"/></xsl:attribute>
                    </xsl:element>
                </xsl:for-each>    
            </xsl:if>
          </xsl:if>  
        </xsl:element>
           
    </xsl:template>
    
    <xsl:template match="as:Rule"><!-- for refactoring of rules -->
       
       <xsl:choose>
           <xsl:when test="$refactorMode='true'">
               <xsl:for-each select="as:constraint">
                   <xsl:choose>
                       <xsl:when test="contains(@action,'nillable')"/><!-- Exclude any nillable rules -->
                       <xsl:otherwise>
                           <xsl:element name="as:constraint">
                               <xsl:for-each select="attribute::*">
                                   <xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute>
                               </xsl:for-each>
                           </xsl:element>  
                           <xsl:copy-of select="as:annotation"/>
                       </xsl:otherwise>
                   </xsl:choose>
               </xsl:for-each>
           </xsl:when>
           <xsl:otherwise><xsl:copy-of select="."/></xsl:otherwise>
       </xsl:choose>

    </xsl:template>
    
    
    <xsl:template match="camed:annotation" mode="rename">
        <xsl:param name="newElements"/>
        <xsl:param name="newAttributes"/>
        
        <xsl:element name="camed:annotation">
                <xsl:for-each select="attribute::*">
                    <xsl:choose>
                        <xsl:when test="name()='item'"><xsl:attribute name="{name()}"><xsl:value-of select="if(starts-with(.,'//')) then '/' else ''"/><xsl:value-of select="camed:chkPath(.,$newElements,$newAttributes)"/></xsl:attribute></xsl:when><!-- lookup path changes -->
                        <xsl:otherwise><xsl:attribute name="{name()}"><xsl:value-of select="."/></xsl:attribute></xsl:otherwise>
                    </xsl:choose>
                </xsl:for-each>
            <xsl:for-each select="camed:documentation"><xsl:copy-of select="."/></xsl:for-each>
        </xsl:element>
    </xsl:template>
    
    <xsl:template match="*" mode="rename"><!-- catch all for any non-matched rename mode structure items -->
        <xsl:copy-of select="."/>
    </xsl:template>
    
    <xsl:template match="*">
        <xsl:copy-of select="."/>
    </xsl:template>
    
    <xsl:template match="as:Extension">
        <xsl:copy-of select="."/>
    </xsl:template>
       
    <xsl:function name="camed:FIX-UPPER">
        <xsl:param name="name"/>
        <xsl:param name="mode"/>
        <xsl:param name="type"/>
        
        <xsl:variable name="lcName" select="lower-case($name)"/>
        
        <xsl:variable name="splitName">
           <xsl:value-of select="camed:splitWords($name,$lcName,'true')"/>
        </xsl:variable>
        <xsl:variable name="splitNameList" select="tokenize($splitName,'\|')"/>
        <xsl:variable name="revisedName">
            <xsl:for-each select="tokenize($splitName,'\|')">
                <xsl:variable name="nextWord" select="position() + 1"/>
                <xsl:value-of select="if(position()=1) then camed:WordDefaults(.,$splitNameList[$nextWord],$mode) else concat('|',camed:WordDefaults(.,$splitNameList[$nextWord],$mode))"/>
            </xsl:for-each>
        </xsl:variable>
        <xsl:variable name="newname"><xsl:value-of select="if($mode='NIEM') then replace($revisedName,'\|','') else replace($revisedName,'\|','-')"/></xsl:variable>
        <xsl:variable name="returnName"><xsl:value-of select="if($type='ELEMENT') then $newname else concat(lower-case(substring($newname,1,1)),substring($newname,2,string-length($newname) - 1))"/></xsl:variable>
        <xsl:value-of select="if($returnName=$name) then '' else $returnName"/>
    </xsl:function>
    
    <xsl:function name="camed:splitWords">
        <xsl:param name="words"/>
        <xsl:param name="lcWords"/>
        <xsl:param name="start"/>
        <xsl:if test="string-length($words) &gt; 0">
            <xsl:variable name="firstChar" select="substring($words,1,1)"/>
            <xsl:variable name="nextChars" select="substring-after($words,$firstChar)"/>
            <xsl:variable name="firstNextChar" select="substring($nextChars,1,1)"/>
            <xsl:choose>
                <xsl:when test="contains('.-_',$firstChar)">|<xsl:value-of select="camed:splitWords($nextChars,lower-case($nextChars),'true')"/></xsl:when>
                <xsl:when test="$start='true'"><xsl:value-of select="upper-case($firstChar)"/><xsl:value-of select="camed:splitWords($nextChars,lower-case($nextChars),'false')"/></xsl:when>
                <xsl:when test="$firstChar=substring($lcWords,1,1)"><xsl:value-of select="$firstChar"/><xsl:value-of select="camed:splitWords($nextChars,lower-case($nextChars),'false')"/></xsl:when>
                <xsl:otherwise><xsl:value-of select="lower-case($firstChar)"/><xsl:value-of select="camed:splitWords($nextChars,lower-case($nextChars),'false')"/></xsl:otherwise>
            </xsl:choose>
        </xsl:if>
    </xsl:function>
    
    <xsl:function name="camed:chkPath">
        <xsl:param name="path"/>
        <xsl:param name="newNames"/>
        <xsl:param name="newAttributes"/>
        <!-- Examine path for matching name changes -->
        <xsl:for-each select="tokenize($path,'/')">
            <xsl:variable name="this" select="."/>
            <xsl:variable name="that" select="if(contains($this,'@')) then substring-after($this,'@') else $this"/>
            <xsl:variable name="prefix" select="if(contains($this,':')) then '' else $insertNSprefix"/>
            <xsl:choose>
                <xsl:when test="$this=''"/><!-- ignore empty tokens -->
                <xsl:when test="contains($this,'@')">/@<xsl:value-of select="if($newAttributes//allAttributes/attribute[@name=$that]) then $newAttributes//allAttributes/attribute[@name=$that]/@newname else $that"/></xsl:when><!-- attributes lookup -->
                <xsl:otherwise>
                    <xsl:variable name="nextPathItem"><xsl:value-of select="if($newNames//allElements/element[@name=$this]) then $newNames//allElements/element[@name=$this]/@newname else $this"/></xsl:variable>
                    <xsl:value-of select="if(contains($nextPathItem,':')) then concat('/',$nextPathItem) else concat('/',$insertNSprefix,$nextPathItem)"></xsl:value-of>
                </xsl:otherwise><!-- elements lookup -->
            </xsl:choose>
        </xsl:for-each>
    </xsl:function>
    
    <xsl:function name="camed:chkExpression">
        <xsl:param name="expression"/>
        <xsl:param name="newNames"/>
        <xsl:param name="newAttributes"/>
        
        <xsl:variable name="path">
            <xsl:choose>
                <xsl:when test="contains($expression,'/')">
                    <xsl:choose>
                        <xsl:when test="contains($expression,'''')"><xsl:value-of select="concat('/',substring-before(substring-after($expression,'/'),''''))"/></xsl:when>
                        <xsl:when test="contains($expression,')')"><xsl:value-of select="concat('/',substring-before(substring-after($expression,'/'),')'))"/></xsl:when>
                        <xsl:otherwise/>
                    </xsl:choose>
                </xsl:when>
                <xsl:otherwise/>
            </xsl:choose>
        </xsl:variable>
        
        <!-- Examine path for matching name changes -->
        <xsl:variable name="newPath">
        <xsl:for-each select="tokenize($path,'/')">
            <xsl:variable name="this" select="."/>
            <xsl:variable name="that" select="if(contains($this,'@')) then substring-after($this,'@') else $this"/>
            <xsl:choose>
                <xsl:when test="$this=''"/><!-- ignore empty tokens -->
                <xsl:when test="contains($this,'@')">/@<xsl:value-of select="if($newAttributes//allAttributes/attribute[@name=$that]) then $newAttributes//allAttributes/attribute[@name=$that]/@newname else $that"/></xsl:when><!-- attributes lookup -->
                <xsl:otherwise>/<xsl:value-of select="if($newNames//allElements/element[@name=$this]) then $newNames//allElements/element[@name=$this]/@newname else $this"/></xsl:otherwise><!-- elements lookup -->
            </xsl:choose>
        </xsl:for-each>
        </xsl:variable>
        
        <!-- Return revised expression -->
        <xsl:choose>
            <xsl:when test="$newPath=''"><xsl:value-of select="$expression"/></xsl:when>
            <xsl:otherwise><xsl:value-of select="concat(substring-before($expression,$path),$newPath,substring-after($expression,$path))"/></xsl:otherwise>
        </xsl:choose>
        
    </xsl:function>
    
</xsl:stylesheet>
